<?php
namespace Ma;

use Ma\Container;
use Ma\Web\Router;
use Ma\Service\Logger;
use Ma\Service\Tpl;

/**
 * @package         Ma
 * @class           App
 * @extends         \Ma\Container
 * @author          Mahdi Abu Yasmine
 * @date            2019-10-15 16:01:18 CET
 * @copyright       meta-tech.academy
 */
abstract class App extends Container
{
    protected string    $name;
    protected array     $services;

    /**
     * @constructor
     * @public
     * @param       string      $name
     * @param       array       $config
     */
    public function __construct(string $name = 'app', array $config = [])
    {
        $this->name     = $name;
        $this->services = [];
        $this->baseInit($config);
    }

    /**
     * @public
     * @return      string
     */
    public function getName():string
    {
        return $this->name;
    }

    /**
     * @public
     * @return      array
     */
    public function availableServices(): array
    {
        return array_keys($this->services);
    }

    /**
     * @public
     * @param       string      $name
     * @return      bool
     */
    public function serviceExists(string $name): bool
    {
        return in_array($name, $this->availableServices());
    }

    /**
     * @public
     * @param       string  $name
     * @param       mixed   $provider
     * @return      bool
     */
    public function registerService(string $name, $provider): bool
    {
        if (($done = !$this->serviceExists($name))) {
            $this->services[$name] = $provider;
        }
        return $done;
    }

    /**
     * @public
     * @param       string  $prefix
     * @param       array   $definition
     * @return      bool
     */
    public function registerServices(string $prefix, array $definition): bool
    {
        $done = true;
        foreach($definition as $name => $obj) {
            $done = $done && $this->registerService("$prefix$name", $obj);
        }
        return $done;
    }

    protected function baseInit(array $config = [])
    {
        $this->services['config'] = $config;
        $this->services['log']    = function(string $name) use (&$logger) {
            if (is_null($logger)) {
                $logger = new Logger(@$this->services['config']['log']);
            }
            return $logger->get($name);
        };
        $this->services['router'] = new Router();
        $this->services['tpl']    = function(?string $name) use (&$tpl) {
            if (is_null($tpl)) {
                $tpl = new Tpl($this->services['config']['tpl']);
            }
            return $tpl->get($name);
        };
        $this->init($config);
    }

    /**
     * @method      init
     * @abstract
     * @public
     * @param       array       $config
     */
    abstract public function init(array $config);

    /**
     * @method      __get
     * @public
     * @param       string      $property
     * @return      mixed
     */
    public function __get(string $property)
    {
        return $this->serviceExists($property) ? $this->services[$property] : null;
    }

    /**
     * @public
     * @static
     * @param      ?string      $name
     * @param       array       $params
     * @return      mixed
     */
    public static function get(?string $name, array $params = [])
    {
        global $core;
        $unit = empty($name) ? $core : $core[$name];
        return is_callable($unit) ? call_user_func_array($unit, $params) : $unit;
    }

    /**
     * @public
     */
    public function run()
    {
        $this->services['router']->resolve($_SERVER["REDIRECT_URL"]);
    }

}
