<?php
namespace Ma\Service;

use Ma\Sys\Dbg;
use Symfony\Component\Yaml\Yaml;
use Exception;
use InvalidArgumentException;

/**
 * @package         Ma\Service
 * @class           Config
 * @author          Mahdi Abu Yasmine
 * @date            2019-08-30 21:22:46 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class ConfigLoader
{
    protected const SALT        = 'x#3mZ5r1Vs7@2:';
    protected const HASH_TYPE   = 'sha256';
    protected const KEY_VAR     = '^vars';
    protected const KEY_IMPORT  = '^imports';
    protected static array $config;
    protected static array $vars;
    protected static array $imports;
    protected static array $map;
    protected static string $rootId;

    /**
     * @private
     * @constructor
     */
    private function __construct()
    {
    }

    /**
     * @method      load
     * @public
     * @static
     * @param       string  $path
     * @param       bool    $root
     * @return      array
     * @throws     \Exception
     */
    public static function load(string $path, bool $root=true): array
    {
        $dtag = Dbg::in(__method__, Dbg::DEBUG, compact(Dbg::args(__method__)), __line__);
        $file = self::getFile($path);
        $id   = self::getId($file);
        self::$imports[$id] = $file;
        if (!isset(self::$config[$id])) {
            if ($root) {
                self::$rootId = $id;
            }
            self::$config[$id] = Yaml::parse(file_get_contents($file));
            self::loadVars($id, $root);
            self::$config[$id] = self::loadEnv(self::$config[$id], self::$vars);
            self::loadImports($id, dirname($file));
        }
//        if ($root) {
            //~ Dbg::dump(self::$map, 'map');
            //~ Dbg::dump(self::$imports, 'imports');
            //~ Dbg::dump(self::$vars, 'vars');
//        }
        Dbg::out($dtag, __line__, self::$config[$id]);
        return self::$config[$id];
    }

    /**
     * @method      getFile
     * @protected
     * @static
     * @param       string  $path
     * @throw      \InvalidArgumentException
     * @return      string
     */
    protected static function getFile(string $path): string
    {
        $file = realpath($path);
        if (!file_exists($file)) {
            throw new InvalidArgumentException("file '$file' do not exists");
        }
        return $file;
    }

    /**
     * @method      loadVars
     * @protected
     * @param       string  $id
     * @param       bool    $root
     */
    protected static function loadVars(string $id, bool $root): void
    {
        $dtag = Dbg::in(__method__, Dbg::DEBUG_4, compact(Dbg::args(__method__)), __line__);
        if ($root) {
            self::$vars = [];
            self::$vars['PATH'] = dirname(self::$imports[$id]);
        }
        if (isset(self::$config[$id][self::KEY_VAR])) {
            $tmp = null;
            foreach (self::$config[$id][self::KEY_VAR] as $key => $val) {
                $tmp = self::loadEnv($val, self::$vars, function($p) { return realpath($p);});
                if (!$tmp) {
                    $tmp = self::loadEnv($val, self::$vars);
                }
                self::$vars[$key] = $tmp;
            }
            unset(self::$config[$id][self::KEY_VAR]);
        }
        Dbg::out($dtag, __line__);
    }

    /**
     * replace string with parameter and/or environment var by related corresponding value
     * * pattern for env var are    : '%UPPERCASE_NAME%'
     * * pattern for var are : '%:UPPERCASE_NAME%'
     *
     * @method      setEnv
     * @private
     * @static
     * @param       string  $str
     * @param       array   $params
     * @return     ?array|?string
     */
    private static function setEnv(string $str, array $params = [])
    {
        if (!empty($params)) {
            $str = preg_replace_callback('/%:([A-Z_][A-Z_\d]*)%/', function($match) use ($params) {
                return $params[$match[1]] ?? $match[1]; 
            }, $str);
        }
        return preg_replace_callback('/%([A-Z_][A-Z_\d]*)%/', function($match) {
            return $_ENV[$match[1]] ?? $match[1]; 
        }, $str);
    }

    /**
     * recursively load environment or config variables in specified $var (string type or [assoc] type)
     *
     * @method      loadEnv
     * @public
     * @statix
     * @param       mixed       $var
     * @param       array       $params
     * @param      ?callable    $apply
     * @return      array|mixed|string|null
     */
    public static function loadEnv($var, array $params=[], ?callable $apply = null)
    {
        if (is_string($var)) {
            $var = is_callable($apply) ? $apply(self::setEnv($var, $params)) : self::setEnv($var, $params);
        }
        elseif (is_array($var) && !empty($var)) {
            $rm = [];
            foreach($var as $k => $v) {
                $nk = self::loadEnv($k, $params, $apply);
                if ($nk != $k) {
                    $rm[] = $k;
                }
                $var[$nk] = self::loadEnv($v, $params, $apply);
            }
            foreach($rm as $k) {
                unset($var[$k]);
            }
        }
        return $var;
    }

    /**
     * @method      loadImports
     * @protected
     * @static
     * @param       string  $id
     * @param       string  $dir
     * @throws     \Exception
     */
    protected static function loadImports(string $id, string $dir)
    {
        if (isset(self::$config[$id][self::KEY_IMPORT])) {
            $tmp = [];
            foreach (self::$config[$id][self::KEY_IMPORT] as $key => $path) {
                if (!isset(self::$config[$id][$key])) {
                    if ($path[0] != '/') {
                        $path = "$dir/$path";
                    }
                    self::setMap($id, $path);
                    self::$config[$id][$key] = self::load($path, false);
                    $tmp = self::loadEnv(self::$config[$id][$key], self::$vars);
                    if (isset($tmp[$key])) {
                        self::$config[$id][$key] = $tmp[$key];
                        foreach($tmp as $k => $v) {
                            if ($k != $key) {
                                self::$config[$id][$key][$k] = $v;
                            }
                        }
                    }
                    else throw new Exception ("Invalid config file content. Imported file must contains the imported key `$key`");
                }
            }
            unset(self::$config[$id][self::KEY_IMPORT]);
        }
    }

    /**
     * @method      setMap
     * @protected
     * @static
     * @param       string  $id
     * @param       string  $path
     * @return      bool
     */
    protected static function setMap(string $id, string $path): bool
    {
        $done = false;
        try {
            self::$map[self::getId(self::getFile($path))] = $id;
            $done = true;
        }
        catch (Exception $e) {
            
        }
        return $done;
    }

    /**
     * @method      getId
     * @protected
     * @static
     * @param       string  $path
     * @return      string
     */
    protected static function getId(string $path): string
    {
        return hash(self::HASH_TYPE, self::SALT . $path);
    }

    /**
     * @method      getVar
     * @public
     * @static
     * @param       string  $name
     * @return     ?string
     */
    public static function getVar(string $name): ?string
    {
        return self::$vars[$name] ?? null;
    }

    /**
     * @method      getVarList
     * @public
     * @static
     * @return      array
     */
    public static function getVarList(): array
    {
        return array_keys(self::$vars);
    }

    /**
     * @method      get
     * @public
     * @static
     * @param       mixed  $name
     * @return      mixed
     */
    public static function get($name)
    {
        $data = null;
        if (isset(self::$rootId)) {
            if (!is_array($name)) {
                if (strpos($name, '.') != -1) {
                    $name = preg_split('/\./', $name);
                }
                else {
                    $name = [$name];
                }
            }
            $d = self::$config[self::$rootId];
            foreach ($name as $n) {
                $d = $d[$n] ?: null;
                if (is_null($d)) break;
            }
            if (!is_null($d) && $d != self::$config[self::$rootId]) {
                $data = $d;
            }
        }
        return $data;
    }
}