<?php
namespace Ma\Sys;

use Ma\Sys\TermFormatter;
use Ma\Sys\TermStyleCollection;
use Ma\Sys\TermStyle;
use Ma\Sys\TimeIt;
use Ma\Sys\Dumper;
use Ma\Sys\MemoryObserver;
use Ma\Util\Swiss;
use ReflectionFunction;
use ReflectionMethod;
use ReflectionException;
use DateTime;

/**
 * @package         Ma\Sys
 * @static
 * @class           Dbg
 * @author          Mahdi Abu Yasmine
 * @date            2019-07-31 15:02:07 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class Dbg
{
    const INLINE              = 0;
    const ENTER               = 1;
    const LEAVE               = 2;
    const ERROR               = 3;
    const WARN                = 4;
    const NOTICE              = 5;
    const INFO                = 6;
    const DEBUG               = 7;
    const DEBUG_1             = 8;
    const DEBUG_2             = 9;
    const DEBUG_3             = 10;
    const DEBUG_4             = 11;
    const LIMIT_STR_VALUE_LEN = 30;
    const LIMIT_OBJ_VALUE_LEN = 20;

    protected static TermFormatter  $of;
    protected static TimeIt         $ti;
    protected static MemoryObserver $mem;
    protected static bool           $active;
    public    static bool           $debug = false;
    protected static int            $level = Dbg::INFO;
    protected static Dumper         $dumper;
    protected static array          $stack = [];
    protected static string         $tag   = '\Ma\Sys\Dbg::GLOBAL_TAG';

    /**
     * @private
     * @constructor
     */
    private function __construct()
    {
    }

    /**
     * initialize logging
     * @method      init
     * @public
     * @static
     * @param      ?string  $path           enable disable log messages
     * @param       string  $forceDebug     debug enable disable log messages
     **/
    public static function init (string $path = null, $forceDebug = false): void
    {
        static::$stack  = [];
        static::$ti     = new TimeIt(6, false, false);
        static::$mem    = new MemoryObserver();
        static::$ti->attach(static::$mem);
        static::$ti->start();
        static::$ti->rec(static::$tag);
        static::$of     = !is_null($path) ? new TermFormatter($path) : new TermStyleCollection();
        static::$active = $forceDebug || static::$of->hasDebug('active');
        static::$level  = static::$of->hasDebug('level');
    }

    /**
     * @method      getExecTime
     * @public
     * @static
     * @return      float
     */
    public static function getExecTime(): float
    {
        static::$ti->rec(static::$tag, true);
        return static::$ti->delta(static::$tag);
    }

    /**
     * @method      getMemoryUsage
     * @public
     * @static
     * @return      int
     */
    public static function getMemoryUsage(): int
    {
        static::$ti->rec(static::$tag, true);
        return static::$mem->delta(static::$tag);
    }

    /**
     * @method      getMemoryPeakUsage
     * @public
     * @static
     * @return      int
     */
    public static function getMemoryPeakUsage(): int
    {
        return memory_get_peak_usage();
    }

    /**
     * @method      getFormatter
     * @public
     * @static
     * @return      \Ma\Sys\TermFormatter
     */
    public static function getFormatter(): TermFormatter
    {
        return static::$of;
    }

    /**
     * get the argument signature for the specified function (or class method) using Reflection
     *
     * @method      args
     * @public
     * @static
     * @param       string      $fn         the method name
     * @param       bool        $addSelf    add this keyword
     * @return      array
     */
    public static function args(string $fn, bool $addSelf=true): array
    {
        $rs = $addSelf ? ['this'] : [];
        try {
            $tmp = explode('::', $fn, 2);
            $f = count($tmp) == 1 ? new ReflectionFunction($fn) : new ReflectionMethod($tmp[0], $tmp[1]);
            foreach ($f->getParameters() as $p) {
                $rs[] = $p->name;
            }
        }
        catch (ReflectionException $e) {
            $rs['exception'] = $e->getMessage();
        }
        return $rs;
    }

    /**
     * get the argument signature for the specified function (or class method) using Reflection
     *
     * @method      argm
     * @public
     * @static
     * @param       string      $fn         the method name
     * @return      array
     */
    public static function argm(string $fn): array
    {
        return self::args($fn, false);
    }

    /**
     * inspect specified function or method $fn
     *
     * @method      inspect
     * @public
     * @static
     * @param       string      $fn     the method name
     * @return      array
     */
    public static function inspect(string $fn): array
    {
        $rs     = [];
        try {
            $tmp = explode('::', $fn, 2);
            $f = count($tmp) == 1 ? new ReflectionFunction($tmp) : new ReflectionMethod($tmp[0], $tmp[1]);
            $path = $f->getFileName();
            $sline = $f->getStartLine() - 1;
            $eline = $f->getEndLine();
            //        $len    = $eline - $sline;
            $data = file($path);
            $data = implode('', array_slice($data, 0, count($data)));
            $data = preg_split("/" . PHP_EOL . "/", $data);
            $source = '';
            for ($i = $sline; $i < $eline; $i++) {
                $source .= "$data[$i]" . PHP_EOL;
            }
            $rs = compact('sline', 'eline', 'source', 'path');
        }
        catch(ReflectionException $e) {
            $rs['exception'] = $e->getMessage();
        }
        return $rs;
    }

    /**
     * @method      dump
     * @public
     * @static
     * @param       mixed       $var
     * @param       string      $varname
     * @return      ?\Ma\Sys\Variable
     */
    public static function dump($var, string $varname='variable'): ?Variable
    {
        if (!isset(static::$dumper)) {
            static::$dumper = new Dumper();
        }
        self::initPrint();
        static::$of->action('Dumping', $varname);
        $data = static::$dumper->inspect($var);
        //~ var_dump($data);
        static::displayVar($data);
        static::$of->state(true);
        self::onPrint();
        return $data;
    }

    /**
     * display specified $var
     *
     * @method      displayVar
     * @protected
     * @static
     * @param       mixed   $var
     * @param       int     $beginIndent
     */
    protected static function displayVar($var, int $beginIndent=5): void
    {
        $indent = str_repeat(' ', $beginIndent);
        //~ if (Dbg::$debug && $var->level == 0) var_dump($var);
        if (!isset($var->type)) {
            return;
        }
        $sym   = in_array($var->type, ['array', 'object', 'recurs']) ? '➤ ' : '⦁ ';
        $stype = in_array($var->type, ['array', 'object', 'recurs']) ? 'symb_list' : 'symb_entry';
        switch($var->type) {
            case 'object' :
                $undefined = $var->id == '#uninitialized';
                printf('%s%s%s%s%s%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    isset($var->visibility) && !is_null($var->visibility) && !is_null($var->parent) && $var->parent->type == 'object' ? static::$of->apply('visibility', substr($var->visibility, 0, 3)).' ' : '',
                    static::$of->apply('list'    , str_pad($var->type, $var->level > 0 ? 8 : 5).' '),
                    $var->level > 0 ? static::$of->apply('object' , $var->name) : '',
                    $var->level > 0 ? static::$of->apply('arg'    , ' : ') : '',
                    static::$of->apply('class' , $var->class),
                    static::$of->apply('hash'   , $var->id),
                    $undefined ? "" : static::$of->apply('bracket'    , ' ('),
                    $undefined ? "" : static::$of->apply('size'    , $var->size),
                    $undefined ? "" : static::$of->apply('bracket'    , ')'),
                    $undefined ? "" : static::$of->apply('prop_sep', ' {')
                );
                echo PHP_EOL;
                if ($undefined) {
                    break;
                }
                foreach($var->data as $v) {
                    static::displayVar($v, $beginIndent);
                }
                echo static::$of->apply('prop_sep', $indent . $var->indent . '}') . PHP_EOL;
                break;
            case 'array' :
                printf('%s%s%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    isset($var->visibility) && !is_null($var->visibility) && !is_null($var->parent) && $var->parent->type == 'object' ? static::$of->apply('visibility', ($var->visibility == '"ArrayObject":private' ? 'pri' : substr($var->visibility, 0, 3))).' ' : '',
                    static::$of->apply('list'    , str_pad($var->type, $var->level > 0 ? 8 : 5)),
                    $var->level > 0 ? static::$of->apply('object'    , ' '.$var->name) : '',
                    static::$of->apply('bracket'    , '['),
                    static::$of->apply('size'    , $var->size ?? 0),
                    static::$of->apply('bracket'    , ']'),
                    static::$of->apply('prop_sep', ' {')
                );
                echo PHP_EOL;
                foreach($var->data as $v) {
                    static::displayVar($v, $beginIndent);
                }
                echo static::$of->apply('prop_sep', $indent . $var->indent . '}') . PHP_EOL;
                break;
            case 'bool' :
            case 'int' :
                printf('%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    isset($var->visibility) && !is_null($var->visibility) && !is_null($var->parent) && $var->parent->type == 'object' ? static::$of->apply('visibility', substr($var->visibility, 0, 3)).' ' : '',
                    static::$of->apply('entry'   , $var->level > 0 ? str_pad($var->type, 8).' ' : $var->type),
                    static::$of->apply('prop' , str_pad($var->name ?? '', isset($var->parent) && $var->parent != null ? $var->parent->maxlen : 0)),
                    static::$of->apply('prop_sep'    , ' : '),
                    static::$of->apply($var->type == 'int' ? 'num' : 'timeit', $var->data)
                );
                echo PHP_EOL;
                break;
            case 'null' :
                printf('%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    !is_null($var->visibility) && !is_null($var->parent) && $var->parent->type == 'object' ? static::$of->apply('visibility', substr($var->visibility, 0, 3)).' ' : '',
                    static::$of->apply('entry'     , $var->level > 0 ? str_pad($var->type, 8).' ' : $var->type),
                    static::$of->apply('prop'      , str_pad($var->name, $var->parent->maxlen)),
                    static::$of->apply('prop_sep'  , ' : '),
                    static::$of->apply('timeit')
                );
                echo PHP_EOL;
                break;
            case 'recurs' :
                printf('%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    static::$of->apply('entry'   , $var->level > 0 ? str_pad($var->parent->type, 8).' ' : $var->type),
                    static::$of->apply('prop' , str_pad($var->name, $var->parent->maxlen)),
                    static::$of->apply('prop_sep'    , ' : '),
                    static::$of->apply('timeit', $var->data),
                    static::$of->apply('hash'  , !is_null($var->parent->id) ? $var->parent->id : '')
                );
                echo PHP_EOL;
                break;
            case 'string' :
                printf('%s%s%s%s%s%s%s%s%s%s',
                    $indent . $var->indent,
                    $var->level > 0 ? static::$of->apply($stype    , $sym) : '',
                    isset($var->visibility) && !is_null($var->visibility) && !is_null($var->parent) && $var->parent->type == 'object' ? static::$of->apply('visibility', substr($var->visibility, 0, 3)).' ' : '',
                    static::$of->apply('entry'     , $var->level > 0  ? str_pad($var->type, 8).' ' : $var->type),
                    isset($var->name) && (!empty($var->name) || $var->name === '0') ? static::$of->apply('prop'    , str_pad($var->name, $var->parent->maxlen)) : '',
                    static::$of->apply('prop_sep'       , ' : '),
                    static::$of->apply('data', '"'.$var->data.'"'),
                    static::$of->apply('bracket'       , ' ('),
                    static::$of->apply('length'       , $var->size),
                    static::$of->apply('bracket'       , ')')
                    
                );
                echo PHP_EOL;
                break;
        }
    }

    /**
     * @method      getStyle
     * @protected
     * @static
     * @param       int         $mode
     * @param      ?string      $method
     * @param      ?string      $caller
     * @return      string
     */
    protected static function getStyle(int $mode, string $method = null, ?string $caller = null): string
    {
        $label = '';
        if (isset(static::$of)) {
            if (!is_null($caller)) {
                if (substr($method, 0, strlen($caller)) == $caller) {
                    $method = substr($method, strlen($caller));
                }
                else {
                    $method = "::$method";
                }
            }
            switch ($mode) {
                case static::ENTER  : 
                    $label = static::$of->apply("enter" , "> $caller").$method; 
                    break;
                case static::LEAVE  : 
                    $label = static::$of->apply("leave" , "< $caller").$method;
                    break;
                case static::ERROR  : 
                    $label = static::$of->apply("error" , " Error: ");
                    break;
                case static::WARN   : 
                    $label = static::$of->apply("warn"  , " Warn: ");
                    break;
                case static::NOTICE : 
                    $label = static::$of->apply("notice", " Notice: ");
                    break;
                default : 
                    $label = "  "; 
                    break;
            }
        }
        return $label;
    }

    /**
     * @method      on
     * @private
     * @static
     * @param       int     $level
     * @return      bool
     */
    private static function on(int $level = self::INFO): bool
    {
        return static::$active && $level <= static::$level;
    }

    /**
     * @method      parseParams
     * @private
     * @static
     * @param      ?array   $params
     * @param       string  $styleKey
     * @param       string  $styleVal
     * @return      string
     */
    private static function parseParams (array $params = null, string $styleKey = 'key', string $styleVal = 'val'): string
    {
        $str = $sep = "";
        if (static::$active && !is_null($params)) {
            foreach($params as $k => $v) {
                $val = $v;
                if (is_string($v) && strlen($v) > static::LIMIT_STR_VALUE_LEN) {
                    $val = substr($v, 0, static::LIMIT_STR_VALUE_LEN).'...';
                }
                elseif (is_array($v) || is_object($v)) {
                    $enc = substr(json_encode($v), 0, static::LIMIT_OBJ_VALUE_LEN);
                    $val = gettype($v) . "=$enc" . (strlen($enc) < static::LIMIT_OBJ_VALUE_LEN ? '' : '...'.(is_array($v) ? ']' : '}'));
                }
                elseif (is_bool($v)) {
                    $val = $v ? 'true' : 'false';
                }
                elseif (is_resource($v)) {
                    $val = "$v";
                }
                $str .= $sep . static::$of->apply($styleKey, $k).':'. static::$of->apply($styleVal, is_null($val) ? 'null' : $val);
                if (empty($sep)) $sep = ', ';
            }
        }
        return $str;
    }

    /**
     * log method entering
     * 
     * @public
     * @param       string      $method         the method name
     * @param       int         $level          the used level
     * @param      ?array       $parameters     optional parameters string to display on entering method
     * @param       int         $line           optional current file line
     * @param      ?string      $file           optional current file
     * @return     ?string
     */
    public static function in(string $method, int $level=self::INFO, ?array $parameters=null, int $line=0, string $file=null): ?string
    {
        $tag = null;
        if (static::on($level)) {
            $tag    = Swiss::genid();
            $caller = isset($parameters['this']) ? get_class($parameters['this']) : $method;
            if ($method == $caller) {
                $tmp    = explode('::', $method, 2);
                $caller = $tmp[0];
                if (count($tmp) == 2) {
                    $method = $tmp[1];
                }
            }
            static::$stack[$tag] = compact('method', 'caller', 'file');
            static::log(sprintf("%s(%s)", static::getStyle(static::ENTER, $method, $caller), static::parseParams($parameters)), $line, static::ENTER, $file, $tag);
            static::$ti->rec($tag);
        }
        return $tag;
    }

    /**
     * log method leaving
     *
     * @method      out
     * @public
     * @param      ?string      $tag        current tag
     * @param       int         $line       optional current file line
     * @param       mixed       $return     optional returned value
     */
    public static function out(string $tag = null, int $line=0, $return = null): void
    {
        if (!empty($tag)) {
            if (static::$of->hasDebug('show_exec_time')) static::$ti->rec($tag);
            extract(static::$stack[$tag]);
            unset(static::$stack[$tag]);
            if (is_string($return)) {
                $return = preg_replace('/' . PHP_EOL . '/', '\\n', $return);
            }
            $r = !is_null($return) ? ' '.static::parseParams(compact('return'), 'return', 'return_val') : '';
            static::log(sprintf("%s(%s)%s", static::getStyle(static::LEAVE, $method, $caller), '', $r), $line, static::LEAVE, $file, $tag, static::$of->hasDebug('show_exec_time') === false  ? null : static::$ti->delta($tag), static::$of->hasDebug('show_memory_usage') === false  ? null : static::$mem->delta($tag));
        }
    }

    /**
     * log a message
     * @method      msg
     * @public
     * @static
     * @param       string      $msg    the msg to log
     * @param       int         $line   optional current file line
     * @param      ?string      $file   optional current file
     **/
    public static function msg(string $msg, int $line = 0, string $file = null): void
    {
        static::log($msg, $line, static::INLINE, $file);
    }

    /**
     * log error
     * @method      error
     * @public
     * @static
     * @param       string  $msg        the error msg to log
     * @param       string  $method     the method name logging the error
     * @param       int     $line       optional current file line
     * @param      ?string  $file       optional current file
     * @param       int     $mode       optional log mode (DBG_ERROR|DBG_WARN)
     **/
    public static function error(string $msg, string $method, int $line = 0, string $file = null, int $mode=Dbg::ERROR): void
    {
        static::log(
            !is_null($file) ? sprintf ("%s %s (in %s)", $msg, TermStyle::off(), $method) : $msg,
            $line,
            $mode,
            is_null($file) ? $method : $file
        );
    }

    /**
     * log a message
     *
     * @method      log
     * @public
     * @static
     * @param       string      $msg        the error msg to log
     * @param       int         $line       optional current file line
     * @param       int         $mode       optional log mode (DBG_ERROR|DBG_WARN)
     * @param      ?string      $file       optional current file
     * @param      ?string      $tag        optional current tag
     * @param      ?float       $deltaTime  optional delta time
     * @param      ?int         $deltaMem   optional delta memory
     **/
    private static function log(string $msg, int $line = 0, int $mode=Dbg::INLINE, string $file = null, string $tag = null, ?float $deltaTime = null, ?int $deltaMem = null): void
    {
        if (static::$active) {
            $time = new DateTime();
            self::initPrint();
            printf (
                "%s%s%s%s%s%s%s%s%s%s %s%s%s%s",
                static::$of->apply ('arg', '['),
                static::$of->hasDebug('show_hours')    === false ? '' : static::$of->apply('date' , $time->format('H:i')),
                static::$of->hasDebug('show_seconds')  === false ? '' : static::$of->apply('time' , $time->format (':s')),
                static::$of->hasDebug('show_micro')    === false ? '' : static::$of->apply('micro', '.' . $time->format('u')),
                !is_null($file) ? static::$of->apply('comment', " $file") : '',
                $line > 0 && ($mode != static::INLINE || !is_null($file)) ? '`'.static::$of->apply('num', sprintf("%04d", $line)) : '',
                static::$of->apply('arg', '] '),
                is_null($tag) || static::$of->hasDebug('show_hash') === false ? '' : static::$of->apply('hash', substr($tag, 0, 7)).' ',
                in_array($mode, [static::ENTER, static::LEAVE]) ? '' : static::getStyle($mode),
                $line > 0 && $mode  == static::INLINE && is_null($file) ? static::$of->apply('num', "$line: "). $msg : $msg,
                !isset($deltaTime) || static::$of->hasDebug('show_exec_time') === false ? '' : static::$of->apply("timeit", " $deltaTime s"),
                !isset($deltaMem) || static::$of->hasDebug('show_memory_usage') === false ? '' : static::$of->apply("memory", " ".Swiss::readsize($deltaMem)),
                TermStyle::off(),
                static::$of->hasDebug('server') ? '' : PHP_EOL
            );
            self::onPrint();
        }
    }

    /**
     * @method      initPrint
     * @private
     * @static
     * @return      void
     */
    private static function initPrint(): void
    {
        if (static::$of->hasDebug('server')) {
            ob_start();
        }
    }

    /**
     * @method      onPrint
     * @private
     * @static
     * @return      void
     */
    private static function onPrint(): void
    {
        if (static::$of->hasDebug('server')) {
            error_log(ob_get_contents());
            ob_end_clean();
        }
    }

}
