<?php
namespace Ma\Sys;

use Ma\Sys\Dbg;
use Ma\Sys\Variable;

/**
 * @package         Ma\Sys
 * @static
 * @class           Dumper
 * @author          Mahdi Abu Yasmine
 * @date            2019-08-01 14:00:44 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class Dumper
{
    protected const NONE            = 0;
    protected const IN              = 1;
    protected const PROP            = 2;
    protected const CONTINUE_STRING = 3;
    protected const END_BLOCK       = 4;
    protected $data;
    protected int $state;
    protected int $level;
    protected ?Variable $profile;

    /**
     * @constructor
     * @public
     */
    public function __construct()
    {
        $this->state = self::NONE;
    }

    /**
     * @method      setData
     * @param       mixed   $var
     * @return      void
     */
    private function setData($var): void
    {
        ob_start();
        var_dump($var);
        $this->data    = preg_split("/".PHP_EOL."/", ob_get_clean());
        $this->state   = self::NONE;
        $this->level   = 0;
        $this->profile = null;
    }

    /**
     * @method      inspect
     * @public
     * @param       mixed       $var
     * @return      \Ma\Sys\Variable
     */
    public function inspect($var): ?Variable
    {
        $this->setData($var);
        $parent = null;
        $v      = null;
        foreach ($this->data as $num => $line) {
            if (Dbg::$debug) {
                echo Dbg::getFormatter()->apply('enter', str_pad($num, 3, '0', STR_PAD_LEFT) . " : $line" . PHP_EOL);
            }
//            var_dump(__line__ . ':' . __method__ . " $num => $line | state : $this->state");
            switch ($this->state) {
                case self::NONE:
                    $this->inspectNone($line, $parent, $v);
                    break;
                case self::IN:
                    $this->inspectInStructure($line, $parent, $v);
                    break;
                case self::END_BLOCK:
                    $this->inspectEndBlock($line, $parent);
                    break;
                case self::CONTINUE_STRING:
                    $this->inspectContinueString($line, $parent, $v);
                    break;
                case self::PROP:
                    $this->inspectInProp($line, $parent, $v);
                    break;
            }
        }
        $this->setMaxLen($this->profile);
        return $this->profile;
    }

    /**
     * @method      inspectInStructure
     * @public
     * @param      ?string              $line
     * @param      ?\Ma\Sys\Variable   &$parent
     * @param      ?\Ma\Sys\Variable   &$v
     * @param       bool               &$bypassCount
     * @return      void
     */
    private function inspectInStructure(?string $line, ?Variable &$parent, ?Variable &$v, bool $bypassCount = false): void
    {
        $isList = isset($parent) && in_array($parent->type, ['object', 'array']);
        if ($isList && count($parent->data) == 0) $this->level++;
        preg_match('/(?P<indent> +)?\["?(?P<name>[^\"]+)"?:?(?P<vis>[^\]]+)?\]=>/', $line, $matches);
        if ($isList && isset($parent) && isset($parent->size) && ($bypassCount || count($parent->data) < $parent->size)) {
            $v = new Variable();
            if (isset($matches['indent'])) {
                if (isset($matches['vis']) && substr($matches['vis'], -strlen(":private")) == ':private') {
                    $matches['vis'] = "priv";
                }
//                var_dump($matches);
                $v->indent = $matches['indent'];
                $v->name = $matches['name'];
                $v->visibility = $matches['vis'] ?? 'public';
                $v->level = $this->level;
                $v->parent = $parent;
                $parent->data[] = $v;
                $this->state = self::PROP;
            }
        }
    }

    /**
     * @method      inspectInStructure
     * @public
     * @param      ?string              $line
     * @param      ?\Ma\Sys\Variable   &$parent
     * @param      ?\Ma\Sys\Variable   &$v
     * @return      void
     */
    private function inspectNone(?string $line, ?Variable &$parent, ?Variable &$v): void
    {
        preg_match('/(?P<indent> +)?(?P<type>[a-z_\d]+)\((?P<name>[^\)]+)\)((?P<id>#\d+) \((?P<length>\d+)\))?/', $line, $matches);
        if (isset($matches['type'])) {
            $this->state = self::IN;
            $v = $this->getVarFromDef($matches, $line);
            $this->profile = $parent = $v;
        }
    }

    /**
     * @method      inspectEndBlock
     * @public
     * @param      ?string              $line
     * @param      ?\Ma\Sys\Variable   &$parent
     * @return      void
     */
    private function inspectEndBlock(?string $line, ?Variable &$parent): void
    {
        if (isset($parent->indent)) {
            $end = substr($line, strlen($parent->indent));
            if (in_array($parent->type, ['object', 'array'])) {
                if ($end == '}') {
                    if ($this->level != 0) {
                        $this->level--;
                        $parent = $parent->parent ?? $parent;
                    }
                    if (isset($parent->data) && $parent->size != count($parent->data)) {
                        $this->state = self::IN;
                    }
                }
                else {
                    $this->inspectInStructure($line, $parent, $v, true);
                }
            }
        }
    }

    /**
     * @method      inspectEndBlock
     * @public
     * @param      ?string              $line
     * @param      ?\Ma\Sys\Variable   &$parent
     * @param      ?\Ma\Sys\Variable   &$v
     * @return      void
     */
    private function inspectInProp(?string $line, ?Variable &$parent, ?Variable &$v): void
    {
        $isNull = $this->state == self::PROP && trim($line) == "NULL";
        $isRecu = $this->state == self::PROP && trim($line) == "*RECURSION*";
        $size = count($parent->data);
        preg_match('/(?P<indent> +)?(?P<type>[a-z_\d]+)\((?P<name>[^\)]+)\)((?P<id>#\d+) \((?P<length>\d+)\)|)?/', $line, $matches);
//        var_dump($line . ' ' . $matches['type']);
        if (isset($matches['type'])) {
            $v = $this->getVarFromDef($matches, $line, $parent->data[$size - 1]);
        }
        elseif ($isNull) {
            $v->type = "null";
            $this->state = self::IN;
        }
        elseif ($isRecu) {
            $v->type = 'recurs';
            $v->data = "*RECURSION*";
            $this->state = self::IN;
        }
        $parent->data[$size - 1] = $v;
        if (in_array($v->type, ['bool', 'int'])) {
            $this->state = self::IN;
        }
        elseif ($v->type == 'string') {
            if (strlen($v->data) == $v->size) {
                $this->state = self::IN;
            }
            else {
                $this->state = self::CONTINUE_STRING;
                return;
            }
        }
        elseif ($v->type == 'object' || $v->type == 'array') {
            if ($matches['type'] == "uninitialized") {
                $parent->size = $parent->size + 1;
            }
            else {
                $parent = $v;
            }
            $this->state = self::IN;
        }
        if (count($parent->data) == $parent->size) {
            $this->state = self::END_BLOCK;
        }
    }

    /**
     * @method      inspectEndBlock
     * @public
     * @param      ?string              $line
     * @param      ?\Ma\Sys\Variable   &$parent
     * @param      ?\Ma\Sys\Variable   &$v
     * @return      void
     */
    private function inspectContinueString(?string $line, ?Variable &$parent, ?Variable &$v)
    {
        $len = strlen($v->data);
        $flen = $v->size - $len - strlen(PHP_EOL);
        $add = $flen > 0 ? PHP_EOL . substr($line, 0, $flen) : PHP_EOL;
        $v->data .= $add;
        if (strlen($v->data) == $v->size) {
            if (isset($parent) && in_array($parent->type, ['object', 'array'])) {
                if (count($parent->data) == $parent->size) {
                    $this->state = self::END_BLOCK;
                }
                else {
                    $this->state = self::IN;
                }
            }
        }
    }
    /**
     * @method      setMaxLen
     * @private
     * @param       ?Variable   &$parent
     */
    private function setMaxLen(?Variable $parent): void
    {
        if (!empty($parent) && isset($parent->type) && in_array($parent->type, ['object', 'array'])) {
            foreach($parent->data as $child) {
                if (empty($parent->maxlen) || $parent->maxlen < strlen($child->name)) {
                    $parent->maxlen = strlen($child->name);
                }
                $this->setMaxLen($child);
            }
        }
    }

    /**
     * @method      getVarFromDef
     * @private
     * @param       array               $rs
     * @param       string              $line
     * @param       ?\Ma\Sys\Variable   $v
     * @return      ?\Ma\Sys\Variable
     */
    private function getVarFromDef(array $rs, string $line, Variable $v = null): ?Variable
    {
//        var_dump(__method__ . " line : $line ".var_export($rs, true));
        if (count($rs) > 1) {
            if (is_null($v)) {
                $v = new Variable();
            }
            $v->indent = $rs['indent'];
            $v->level  = $this->level;
            $v->type   = $rs['type'];
            if ($v->type == 'object') {
                $v->class = $rs['name'];
                $v->id    = $rs['id'];
                $v->size  = $rs['length'];
                $v->data  = [];
            }
            else {
                if ($v->type == 'array') {
                    $v->data = [];
                    $v->size = $rs['name'];
                }
                elseif (in_array($v->type, ['bool', 'int'])) {
                    $v->data = $rs['name'];
                }
                elseif ($v->type == 'uninitialized') {
                    $v->class = $rs['name'];
                    $v->id    = "#uninitialized";
                    $v->size  = 0;
                    $v->data  = [];
                    $v->type = 'object';
                }
                else {
                    $v->size = $rs['name'];
                }
                if ($v->type == 'string') {
                    $v->data = substr($line, strlen($v->indent.$v->type.'('.$v->size.') "'), $v->size);
                    if (strlen($v->data) != $v->size) {
                        $this->state = self::CONTINUE_STRING;
                    }
                }
            }
        }
        return $v;
    }
}
