<?php
namespace Ma\Sys;

use Ma\Sys\TermOption;
use Ma\Sys\TermStyle;
use Ma\Sys\TermStyleCollection;

/**
 * @package         Ma\Sys
 * @static
 * @class           TermFormatter
 * @author          Mahdi Abu Yasmine
 * @date            2019-07-31 11:52:52 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class TermFormatter
{
    const SPACE  = ' ';
    const SEP    = "─";
    const SYMBOL = "➤";

    protected TermStyleCollection $styles;

    /**
     * @constructor
     * @public
     * @param       ?string     $path
     */
    public function __construct (?string $path)
    {
        $this->styles = empty($path) ? new TermStyleCollection() : TermStyleCollection::newFromYamlFile($path);
    }

    /**
     * get terminal preferred width display
     *
     * @method      width
     * @public
     * @return      int
     */
    public function width(): int
    {
        return $this->getConfig('default_width');
    }

    /**
     * get terminal preferred height display
     *
     * @method      height
     * @public
     * @return      int
     */
    public function height(): int
    {
        return $this->getConfig('default_height');
    }

    /**
     * get used TermStyleCollection
     *
     * @method      getStyles
     * @public
     * @return      \Ma\Sys\TermStyleCollection
     */
    public function getStyles(): TermStyleCollection
    {
        return $this->styles;
    }

    /**
     * get specified $key in config
     *
     * @method      getConfig
     * @public
     * @param       string $key
     * @return      mixed
     */
    public function getConfig(string $key)
    {
        return $this->styles->getConfig($key);
    }

    /**
     * get specified $key in debug config
     *
     * @method      hasDebug
     * @public
     * @param       string      $key
     * @return      mixed
     * @return     ?bool
     */
    public function hasDebug(string $key): ?bool
    {
        return $this->styles->getConfig('debug')[$key] ?? null;
    }

    /**
     * get title separator string
     *
     * @method      titleSep
     * @public
     * @return      string
     */
    public function titleSep(): string
    {
        return sprintf(" %s%s%s", $this->styles->apply("sep", "", false), str_repeat(self::SEP, $this->width()), TermStyle::off());
    }

    /**
     * display a title
     *
     * @method      title
     * @public
     * @param       string      $label      generally title of program
     * @param      ?string      $version    version of program
     * @param      ?string      $author     author of program
     * @param      ?string      $license    license of program
     */
    public function title(string $label, string $version=null, string $author=null, string $license=null): void
    {
        $s    = '';
        $size = strlen($label) + 10;
        if (!is_null($version)) {
            $s .= $this->apply("title_item", " v$version");
            $size += strlen($version) + 2;
        }
        if (!is_null($author)) {
            $s .= $this->apply("title", " | by ")
                . $this->apply("title_item", $author);
            $size += strlen($author) + 6;
        }
        if (!is_null($license)) {
            $size += strlen($license) + 11;
        }
        printf (
            "\n%s\n  %s%s%s%s%s%s\n%s\n",
            $this->titleSep(),
            $this->apply("title", "  $label"),
            $s,
            $this->apply("title", "  "),
            $this->strIndent($this->width() - $size),
            $license != null ? $this->apply ("license", "license : $license") : "",
            TermStyle::off (),
            $this->titleSep ()
        );
    }

    /**
     * display an action
     *
     * @method      action
     * @public
     * @param       string      $name   name of current action
     * @param       string      $val    optional action parameter
     */
    public function action(string $name, string $val=""): void
    {
        printf (
            "%s%s %s %s %s%s",
            PHP_EOL,
            $this->strIndent(-2),
            $this->apply("item"  , self::SYMBOL),
            $this->apply("action", $name),
            $this->apply("val"   , $val),
            PHP_EOL
        );
    }

    /**
     * display a warning
     *
     * @method      warn
     * @public
     * @param       string      $label  warn message to display
     */
    public function warn(string $label): void
    {
        $this->error($label, true);
    }

    /**
     * display an error
     *
     * @method      error
     * @public
     * @param       string      $label  the error message to display
     * @param       bool        $warn   warn flag
     */
    public function error(string $label, bool $warn = false): void
    {
        fwrite(STDERR, 
            sprintf (
                "%s%s%s%s",
                PHP_EOL,
                $this->strIndent(-2),
                $this->apply($this->test(!$warn, true), sprintf ("  %s : %s  ", $warn ? "warn" : "error", $label)),
                PHP_EOL
            )
        );
    }

    /**
     * @method      test
     * @protected
     * @param       bool    $exp
     * @param       bool    $error
     * @return      string
     */
    protected function test(bool $exp, bool $error = false): string
    {
        return !$error ? ($exp ? "done" : "fail") : ($exp ? "error" : "warn");
    }

    /**
     * display a test result
     *
     * @method      rs
     * @public
     * @param       bool    $test   the result of test
     * @param       string  $done   the done label to display if test succeed
     * @param       string  $fail   the fail label to display if test fail
     */
    public function rs(bool $test, string $done = "done", string $fail = "fail"): void
    {
        printf (
            "%s%s%s%s",
            PHP_EOL,
            $this->strIndent(-2),
            $this->apply($this->test($test), sprintf ("  %s  ", $test ? $done : $fail)),
            PHP_EOL
        );
    }

    /**
     * display a state for specified test
     *
     * @method      state
     * @public
     * @param       bool    $test   the test result
     */
    public function state(bool $test): void
    {
        printf (
            "%s%s %s%s", 
            $this->strIndent(),
            $this->apply("sep", str_repeat('-', $this->width() - 16)),
            $this->apply($this->test($test), sprintf ("  %s  ", $test ? "OK" : "KO")),
            PHP_EOL
        );
    }

    /**
     * display a key value message
     *
     * @method      keyval
     * @public
     * @param       string  $key        the key name
     * @param       string  $val        the value
     * @param       string  $styleKey   the style to apply to key content
     * @param       string  $styleVal   the style to apply to value content
     */
    public function keyval(string $key, string $val = "", string $styleKey = "key", string $styleVal = "val")
    {
        $len = $this->styles->getConfig('key_maxlen') - strlen($key);
        printf (
            "%s      %s %s  : %s%s",
            $this->strIndent(),
            $this->apply($styleKey, $key),
            str_repeat(" ", $len),
            $this->apply($styleVal, !is_null($val) ? $val : ""),
            PHP_EOL
        );
    }

    /**
     * generic method to display messages
     *
     * @method      echo
     * @public
     * @param       string  $data           the message to display
     * @param       bool    $linefeed       enable line feed
     * @param       bool    $indentAll      enable multiple line indentation
     * @param      ?string  $styleName      the style to apply
     * @param       int     $indentAdjust   indentation adjust
     */
    public function echo(string $data="", bool $linefeed=true, bool $indentAll=false, string $styleName=null, int $indentAdjust = 0): void
    {
        $s = '';
        if ($indentAll) {
            $begin = true;
            foreach (preg_split('/'.PHP_EOL.'/', $data) as $line) {
                $s .= ($begin ? "" : sprintf("%s%s", PHP_EOL, $this->strIndent($indentAdjust))) . $line;
                if ($begin) $begin = false;
            }
        }
        else $s = $data;
        printf(
            "%s%s%s", 
            $this->strIndent($indentAdjust),
            $this->apply($styleName, $s), 
            $linefeed ? PHP_EOL : ""
        );
    }

    /**
     * display usage command
     * 
     * @public
     * @param       string  $bin        the binary name
     * @param       string  $command    the command name
     * @param       string  $param      the optional parameter
     * @param      ?string  $comment    the optional comment
     */
    public function usageCommand(string $bin, string $command, string $param="", string $comment=null): void
    {
        if ($comment != null) {
            $this->echo($comment, true, true, "comment", 4);
        }
        $this->echo(
            sprintf (
                "%s  %s  %s",
                $bin,
                $this->apply ("command", $command),
                $this->apply ("param"  , $param)
            ), true, false, "bin", 4
        );
    }

    /**
     * display usage option
     *
     * @method      usageOption
     * @public
     * @param      ?string  $name       the option name
     * @param      ?string  $shortname  the option shortname
     * @param      ?string  $argname    the option argument
     * @param       bool    $val        specified optional value argument
     * @param       int     $echoUp     line to move up on displaying option definition
     */
    public function usageOption (string $name=null, string $shortname=null, string $argname=null, bool $val=false, int $echoUp=0): void
    {
        $opt = $arg = "";
        if (!is_null($shortname)) {
            $opt = sprintf (
                "%s%s%s",
                $this->apply('option_sep', '-'),
                $this->apply('option', $shortname),
                !is_null($name) ? $this->apply ('option_sep', ', --') : ''
            );
        }
        if (!is_null($name)) {
            $opt .= $this->apply ('option', $name);
        }
        if (!is_null($argname)) {
            $arg = sprintf (
                "  %s%s%s",
                !$val ? $this->apply('arg_sep', '[') : '',
                $this->apply('arg', $argname),
                !$val ? $this->apply('arg_sep', ']') : ''
            );
        }
        if ($echoUp > 0) {
            $this->echoUp($echoUp, sprintf ("%s%s%s", $this->strIndent(), $opt, $arg));
        }
        else {
            printf("%s%s%s%s", $this->strIndent(), $opt, $arg, PHP_EOL);
        }
    }

    /**
     * display a message up tp x lines
     *
     * @method      echoUp
     * @public
     * @param       int     $line   number of line to move up
     * @param       string  $msg    the message to display
     */
    public function echoUp(int $line, string $msg): void
    {
        printf ("%s\r%s%s\r", $this->moveUp($line), $msg, $this->moveDown($line));
    }

    /**
     * get ansi escape sequence moving up lines
     *
     * @method      moveUp
     * @public
     * @param       int     $line   number of line to move up
     * @return      string
     */
    public function moveUp(int $line): string
    {
        return sprintf("\033[%dA", $line);
    }

    /**
     * get ansi escape sequence moving down lines
     *
     * @method      moveDown
     * @public
     * @param       int     $line   number of line to move down
     * @return      string
     */
    public function moveDown(int $line): string
    {
        return sprintf ("\033[%dB", $line);
    }

    /**
     * get an indentation only string
     *
     * @method      apply
     * @public
     * @param      ?string      $style
     * @param       string      $content
     * @param       bool        $off
     * @return      string
     */
    public function apply(?string $style, string $content="", bool $off=true): string
    {
        return $this->styles->apply($style, $content, $off);
    }

    /**
     * get an indentation only string
     *
     * @method      strIndent
     * @public
     * @param       int     $adjust     adjust number of space indentation adjustment
     * @return      string
     */
    public function strIndent(int $adjust=0):string
    {
        return str_repeat(self::SPACE, $this->styles->getConfig('indent')+$adjust);
    }

    /**
     * @method      format
     * @public
     * @param       string      $label
     * @param      ?string      $styleName
     * @param       bool        $linefeed
     * @param       bool        $offColor
     * @return      string
     */
    private function format(string $label, string $styleName=null, bool $linefeed=true, bool $offColor=false):string
    {
        return sprintf (
            "%s%s%s%s",
            $this->strIndent(),
            $this->apply($styleName, $label),
            $offColor ? TermStyle::off () : "",
            $linefeed ? PHP_EOL : ""
        );
    }

    /**
     * @method      oprint
     * @public
     * @param       string      $label
     * @param      ?string      $styleName
     * @param       bool        $linefeed
     * @param       bool        $offColor
     */
    public function oprint(string $label, string $styleName=null, bool $linefeed=true, bool $offColor=false): void
    {
        echo $this->format($label, $styleName, $linefeed, $offColor);
    }

    /**
     * @method      eprint
     * @public
     * @param       string      $label
     * @param      ?string      $styleName
     * @param       bool        $linefeed
     * @param       bool        $offColor
     */
    public function eprint(string $label, string $styleName=null, bool $linefeed=true, bool $offColor = true): void
    {
        fwrite(STDERR, $this->format($label, $styleName, $linefeed, $offColor));
    }
}
