<?php
namespace Ma\Sys;

use Ma\Sys\TermOption;
use Ma\Sys\TermStyle;
use Symfony\Component\Yaml\Yaml;
use InvalidArgumentException;

/**
 * @package         Ma\Sys
 * @static
 * @class           TermStyle
 * @author          Mahdi Abu Yasmine
 * @date            2019-07-24 22:59:10 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class TermStyleCollection
{
    protected array $styles;
    protected array $config;

    /**
     * @public
     * @constructor
     * @param       array   $config
     */
    public function __construct(array $config = ['term' => ['config' =>[], 'styles' => null]])
    {
        $this->styles = [];
        $this->config = [];
        $this->parseConfig($config);
    }

    /**
     * @public
     * @static
     * @param       string      $path
     * @return      \Ma\Sys\TermStyleCollection
     */
    public static function newFromYamlFile(string $path): TermStyleCollection
    {
        if (!file_exists($path)) {
            throw new InvalidArgumentException("file '$path' do not exists");
        }
        $class  = static::class;
        $config = Yaml::parse(file_get_contents($path));
        return new $class($config);
    }

    /**
     * @method      parseConfig
     * @private
     * @static
     * @param       array   $config
     */
    private function parseConfig(array $config)
    {
        if (isset($config['term']) && isset($config['term']['config']) && isset($config['term']['styles'])) {
            foreach($config['term']['config'] as $k => $v) {
                if (!in_array($k, ['indent', 'key_maxlen', 'default_width', 'default_height']) || !is_int($v)) {
                    throw new InvalidArgumentException("invalid config key $k or value $v in config");
                }
            }
            $this->config = $config['term']['config'];
            $allowedKeys  = ['server', 'active', 'show_hours', 'show_seconds', 'show_micro', 'show_exec_time', 'show_memory_usage', 'show_hash', 'level'];
            foreach($config['term']['debug'] as $k => $v) {
                if (!in_array($k, $allowedKeys) || (!is_bool($v) && $k != 'level')) {
                    throw new InvalidArgumentException("invalid debug key $k or value $v in config");
                }
            }
            $this->config['debug'] = $config['term']['debug'];
            foreach($config['term']['styles'] as $name => $def) {
                $fg    = isset($def['fg'])  ? new TermColor($def['fg'][0], $def['fg'][1], $def['fg'][2]) : null; 
                $bg    = isset($def['bg'])  ? new TermColor($def['bg'][0], $def['bg'][1], $def['bg'][2]) : null; 
                $opt   = [];
                if (isset($def['opt'])) {
                    foreach($def['opt'] as $o) {
                        $opt[] = is_int($o) ? $o : constant(TermOption::class . "::" . strtoupper($o));
                    }
                }
                $this->setStyle($name, new TermStyle($fg, $bg, $opt));
            } 
        }
        else throw new InvalidArgumentException("bad config parameter, missing keys");
    }

    /**
     * @public
     * @return      array
     */
    public function getAvailableStyles(): array
    {
        return array_keys($this->styles);
    }

    /**
     * @method      getConfig
     * @public
     * @param       string      $key
     * @return      mixed
     */
    public function getConfig(string $key)
    {
        return $this->config[$key] ?? null;
    }

    /**
     * @public
     * @param      ?string      $name
     * @param       bool        $applied
     * @return      mixed
     */
    public function getStyle(?string $name, bool $applied=false)
    {
        $style = !is_null($name) && isset($this->styles[$name]) ? $this->styles[$name] : null;
        return $applied ? (!is_null($style) ? $style->apply() : '') : $style;
    }

    /**
     * @method      setStyle
     * @public
     * @param       string      $name
     * @param       \Ma\Sys\TermStyle
     */
    public function setStyle(string $name, TermStyle $style): void
    {
        $this->styles[$name] = $style;
    }

    /**
     * @method      apply
     * @public
     * @param      ?string      $styleName
     * @param       string      $content
     * @param       bool        $off
     * @return      string
     */
    public function apply(string $styleName=null, string $content="", bool $off = true): string
    {
        return sprintf(
            "%s%s%s", 
            $this->getStyle($styleName, true), 
            !is_null($content) ? $content : '',
            $off ? TermStyle::off () : ""
        );
    }
}
