<?php
namespace Ma\Sys;

use Ma\Util\Observable;
use LogicException;
use InvalidArgumentException;

/**
 * @package         Ma\Sys
 * @class           TimeIt
 * @author          Mahdi Abu Yasmine
 * @date            2019-07-24 17:29:22 CET
 * @copyright       meta-tech.academy
 * @license         GNU GPL v3
 */
class TimeIt extends Observable
{
    public const MAIN_TAG  = '_timeit';
    public const KEY_START = 0;
    public const KEY_END   = 1;
    public const KEY_TAG   = 0;
    public const KEY_STATE = 1;
    protected int       $digit;
    protected bool      $defRem;
    protected bool      $strict;
    protected array     $data;
    protected array     $lastRec;

    /**
     * @constructor
     * @public
     * @param       int     $precision
     * @param       bool    $strict
     * @param       bool    $defaultRemove
     */
    public function __construct(int $precision=6, bool $strict=true, bool $defaultRemove=true)
    {
        parent::__construct();
        $this->digit  = min($precision, 22);
        $this->defRem = $defaultRemove;
        $this->strict = $strict;
        $this->data   = [];
    }

    /**
     * @method      startIfNeeded
     * @private
     */
    private function startIfNeeded()
    {
        if (!isset($this->data[self::MAIN_TAG])) $this->start();
    }

    /**
     * @method      start
     * @public
     */
    public function start(): void
    {
        $this->data = [];
        $this->setData(self::MAIN_TAG, self::KEY_START, microtime(true));
    }

    /**
     * @method      rec
     * @public
     * @param       string      $name
     * @param      ?bool        $remove
     */
    public function rec(string $name, ?bool $remove=null): void
    {
        $this->startIfNeeded();
        if ($name == self::MAIN_TAG) {
            if (!$this->strict) return;
            throw new InvalidArgumentException("TimeIt::rec('$name') : invalid \$name parameter value, must not be '".self::MAIN_TAG."'");
        }
        $k = intval(isset($this->data[$name]));
        if ($k == 1 && ($remove || (is_null($remove) && $this->defRem))) {
            unset($this->data[$name]);
            $k = self::KEY_START;
        }
        $this->setData($name, $k, microtime(true));
    }

    /**
     * @private
     * @param       string      $name
     * @param       int         $state
     * @param       float       $value
     */
    private function setData(string $name, int $state, float $value): void
    {
        $this->data[$name][$state] = $value;
        $this->lastRec = [$name, $state];
        $this->notify();
    }

    /**
     * @method      getLastState
     * @public
     * @public
     * @return      int
     */
    public function getLastState(): int
    {
        return $this->lastRec[self::KEY_STATE];
    }

    /**
     * @method      getLastTag
     * @public
     * @return      string
     */
    public function getLastTag(): string
    {
        return $this->lastRec[self::KEY_TAG];
    }

    /**
     * @method      getLastValue
     * @public
     * @return      float
     */
    public function getLastValue(): float
    {
        return $this->data[$this->getLastTag()][$this->getLastState()];
    }

    /**
     * @method      delta
     * @public
     * @param      ?string      $name
     * @return     ?float
     */
    public function delta(?string $name): ?float
    {
        $delta = null;
        if (is_null($name)) $name = self::MAIN_TAG;
        if (!empty($this->data[$name])) {
            $size     = count($this->data[$name]);
            $startTag = $size == 2 ? $name : self::MAIN_TAG;
            $endIndex = $size == 2 ? self::KEY_END : self::KEY_START;
            $delta    = number_format((($this->data[$name][$endIndex]-$this->data[$startTag][self::KEY_START])), $this->digit);
        }
        return $delta;
    }

    /**
     * @method      stop
     * @public
     */
    public function stop(): void
    {
        if (!isset($this->data[self::MAIN_TAG])) {
            if (!$this->strict) return;
            throw new LogicException("TimeIt::stop() : cannot stop if not started first");
        }
        $this->setData(self::MAIN_TAG, self::KEY_END, microtime(true));
    }
}
